"""Echo-Morse: A Python tool for converting text to audible Morse code using various voices."""

__version__ = "0.8.0"

from .translator import translator
from .synthesizer import generate_morse_audio
# Uncomment when implemented
# from .interpreter import interpreter 

__all__ = ['translator', 'generate_morse_audio'] 


import pytest
import os
import sys
import tempfile
import shutil
import random  # Ensure random is imported for mocking its methods
import json  # For creating dummy voice_config.json if needed by voice_manager mocks
from unittest.mock import patch, MagicMock, call

# Add src directory to path
SRC_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "../src"))
sys.path.insert(0, SRC_DIR)

from echomorse.synthesizer import synthesizer  # This is the module under test
from echomorse.utils import voice_manager  # To mock its functions
from pydub import AudioSegment

# Morse timing constants from synthesizer to verify calculated durations
DEFAULT_UNIT_DURATION_MS = synthesizer.DEFAULT_UNIT_DURATION_MS
SYMBOL_SPACE_UNITS = synthesizer.SYMBOL_SPACE_UNITS
LETTER_SPACE_UNITS = synthesizer.LETTER_SPACE_UNITS
WORD_SPACE_UNITS = synthesizer.WORD_SPACE_UNITS
DEFAULT_WPM = synthesizer.DEFAULT_WPM


@pytest.fixture
def test_output_file():
    """Fixture for test output file."""
    # Use tempfile for the output to avoid clutter and ensure cleanup
    fd, path = tempfile.mkstemp(suffix=".wav")
    os.close(fd)  # Close the file descriptor immediately
    yield path
    if os.path.exists(path):
        os.remove(path)


@pytest.fixture
def temp_voice_pack_dir():
    """Creates a temporary directory for a voice pack and cleans up afterward."""
    path = tempfile.mkdtemp(prefix="echomorse_test_voice_")
    yield path
    shutil.rmtree(path)


@pytest.fixture
def dummy_audio_file_creator(temp_voice_pack_dir):
    """Fixture to create a dummy WAV file in the temp_voice_pack_dir."""

    def _creator(filename: str, duration_ms: int = 100) -> str:
        filepath = os.path.join(temp_voice_pack_dir, filename)
        AudioSegment.silent(duration=duration_ms).export(filepath, format="wav")
        return filepath

    return _creator


@pytest.fixture
def voice_config_with_sequences():
    """Create a voice config with sequence patterns."""
    return {
        "name": "test_voice_with_sequences",
        "elements": {
            "dot": "dot.wav",
            "dash": "dash.wav",
            "sos": "sos.wav",
            "hello": "hello.wav",
        },
        "sequence_patterns": {"sos": "... --- ...", "hello": ".... . .-.. .-.. ---"},
    }


class TestSynthesizer:
    def test_generate_default_tone(self):
        """Test the generate_default_tone function."""
        tone = synthesizer.generate_default_tone(duration_ms=150, frequency=440)
        assert isinstance(tone, AudioSegment)
        assert len(tone) == 150
        assert tone.channels == 1
        assert tone.frame_rate == synthesizer.SAMPLE_RATE
        # Could also check fade in/out if pydub allows inspection, but length/type is good start

    def test_load_audio_element_single_file(
        self, temp_voice_pack_dir, dummy_audio_file_creator
    ):
        """Test _load_audio_element with a single valid filename."""
        # Create a real test file to avoid path not found issues
        test_file = "test.wav"
        test_path = dummy_audio_file_creator(test_file)

        result = synthesizer._load_audio_element(
            test_file, temp_voice_pack_dir, "test_el",
            fade_type="percentage", fade_value=5.0, target_dbfs=-20.0
        )
        assert isinstance(result, AudioSegment)
        assert len(result) == 100  # The default duration from dummy_audio_file_creator

    def test_load_audio_element_list_of_files(
        self, temp_voice_pack_dir, dummy_audio_file_creator
    ):
        """Test _load_audio_element with a list of filenames."""
        # Create real test files to avoid path not found issues
        file1 = "file1.wav"
        file2 = "file2.wav"
        file3 = "file3.wav"

        dummy_audio_file_creator(file1, 100)
        dummy_audio_file_creator(file2, 110)
        dummy_audio_file_creator(file3, 120)

        file_list = [file1, file2, file3]

        # Mock random.choice to return a predictable file
        with patch("random.choice", return_value=file2) as mock_random_choice:
            result = synthesizer._load_audio_element(
                file_list, temp_voice_pack_dir, "test_el_list",
                fade_type="percentage", fade_value=5.0, target_dbfs=-20.0
            )

            mock_random_choice.assert_called_once_with(file_list)
            assert isinstance(result, AudioSegment)
            assert len(result) == 110  # Length of file2

    def test_load_audio_element_file_not_found(self, temp_voice_pack_dir):
        """Test _load_audio_element when the audio file does not exist."""
        result = synthesizer._load_audio_element(
            "nonexistent.wav", temp_voice_pack_dir, "test_el_missing",
            fade_type="percentage", fade_value=5.0, target_dbfs=-20.0
        )
        assert result is None

    def test_load_audio_element_invalid_type(self, temp_voice_pack_dir):
        """Test _load_audio_element with an invalid element_value type."""
        result = synthesizer._load_audio_element(
            123, temp_voice_pack_dir, "test_el_invalid_type",
            fade_type="percentage", fade_value=5.0, target_dbfs=-20.0
        )  # type: ignore
        assert result is None
        result = synthesizer._load_audio_element(
            [], temp_voice_pack_dir, "test_el_empty_list",
            fade_type="percentage", fade_value=5.0, target_dbfs=-20.0
        )
        assert result is None

    def test_load_audio_element_load_exception(
        self, temp_voice_pack_dir, dummy_audio_file_creator
    ):
        """Test error handling when AudioSegment.from_file raises an exception."""
        # Create a corrupt WAV file for testing
        corrupt_file = "corrupt.wav"
        path = os.path.join(temp_voice_pack_dir, corrupt_file)
        with open(path, "w") as f:
            f.write("This is not a valid WAV file")

        # Test the error handling
        with patch("logging.Logger.error") as mock_logger:
            result = synthesizer._load_audio_element(
                corrupt_file, temp_voice_pack_dir, "corrupt_el",
                fade_type="percentage", fade_value=5.0, target_dbfs=-20.0
            )
            assert result is None
            mock_logger.assert_called_once()
            # Check that the error message contains the expected parts
            error_msg = mock_logger.call_args[0][0]
            assert "Error loading audio file" in error_msg
            assert corrupt_file in error_msg

    def test_get_basic_morse_elements_with_voice(
        self, temp_voice_pack_dir, dummy_audio_file_creator
    ):
        """Test get_basic_morse_elements with a voice config."""
        dot_path = dummy_audio_file_creator("dot.wav", 100)
        dash_path = dummy_audio_file_creator("dash.wav", 300)

        voice_config = {
            "name": "test_voice",
            "elements": {"dot": "dot.wav", "dash": "dash.wav"},
        }

        actual_unit_time_ms = 120  # WPM adjusted time

        dot_audio, dash_audio = synthesizer.get_basic_morse_elements(
            voice_config, temp_voice_pack_dir, actual_unit_time_ms,
            fade_type="percentage", fade_value=5.0, target_dbfs=-20.0
        )

        # Check that we got AudioSegments back
        assert isinstance(dot_audio, AudioSegment)
        assert isinstance(dash_audio, AudioSegment)

        # Check the durations
        assert len(dot_audio) == 100  # Original dot.wav length
        assert len(dash_audio) == 300  # Original dash.wav length

    def test_get_basic_morse_elements_default_fallback(self):
        """Test get_basic_morse_elements with no voice config."""
        actual_unit_time_ms = 120  # WPM adjusted time

        dot_audio, dash_audio = synthesizer.get_basic_morse_elements(
            None, None, actual_unit_time_ms,
            fade_type="percentage", fade_value=5.0, target_dbfs=-20.0
        )

        # Check that we got AudioSegments back
        assert isinstance(dot_audio, AudioSegment)
        assert isinstance(dash_audio, AudioSegment)

        # Check the durations
        assert len(dot_audio) == actual_unit_time_ms
        assert len(dash_audio) == actual_unit_time_ms * 3  # Default dash is 3x dot

    def test_get_basic_morse_elements_partial_fallback(
        self, temp_voice_pack_dir, dummy_audio_file_creator
    ):
        """Test get_basic_morse_elements with a voice config missing some elements."""
        # Create only a dot.wav, no dash.wav
        dot_path = dummy_audio_file_creator("dot.wav", 100)

        voice_config = {
            "name": "test_voice_partial",
            "elements": {
                "dot": "dot.wav"
                # No dash element
            },
        }

        actual_unit_time_ms = 120  # WPM adjusted time

        dot_audio, dash_audio = synthesizer.get_basic_morse_elements(
            voice_config, temp_voice_pack_dir, actual_unit_time_ms,
            fade_type="percentage", fade_value=5.0, target_dbfs=-20.0
        )

        # Check that we got AudioSegments back
        assert isinstance(dot_audio, AudioSegment)
        assert isinstance(dash_audio, AudioSegment)

        # Check the durations
        assert len(dot_audio) == 100  # Original dot.wav length
        assert len(dash_audio) == actual_unit_time_ms * 3  # Default dash duration

    # --- Tests for generate_morse_audio ---

    @patch.object(synthesizer.voice_manager, "get_voice_dir_path")
    @patch.object(synthesizer.voice_manager, "load_voice_config")
    @patch(
        "echomorse.synthesizer.synthesizer._load_audio_element"
    )  # Patch the helper directly
    def test_gma_voice_basic_dot_dash(
        self,
        mock_load_element,
        mock_load_cfg,
        mock_get_dir,
        test_output_file,
        temp_voice_pack_dir,
        dummy_audio_file_creator,
    ):
        """Test with a voice providing basic dot and dash files."""
        voice_name = "testvoice_basic"
        dot_path = dummy_audio_file_creator("v_dot.wav", 100)
        dash_path = dummy_audio_file_creator("v_dash.wav", 300)

        mock_load_cfg.return_value = {
            "name": voice_name,
            "elements": {"dot": "v_dot.wav", "dash": "v_dash.wav"},
        }
        mock_get_dir.return_value = temp_voice_pack_dir

        mock_dot_segment = AudioSegment.silent(duration=100)
        mock_dash_segment = AudioSegment.silent(duration=300)

        # _load_audio_element will be called for 'dot', then 'dash' by get_segment (pre-caching)
        # then potentially again during morse processing if not using cache correctly (which it should)
        def side_effect_load_element(element_value, voice_dir, el_key, fade_type="percentage", fade_value=5.0, target_dbfs=-20.0):
            if el_key == "dot":
                return mock_dot_segment
            if el_key == "dash":
                return mock_dash_segment
            return None

        mock_load_element.side_effect = side_effect_load_element

        synthesizer.generate_morse_audio(
            ".-", test_output_file, voice_name_full=voice_name, wpm=20,
            fade_type="percentage", fade_value=5.0, target_dbfs=-20.0
        )

        # Check a dot and dash were loaded
        mock_load_element.assert_any_call(
            "v_dot.wav", temp_voice_pack_dir, "dot", "percentage", 5.0, -20.0
        )
        mock_load_element.assert_any_call(
            "v_dash.wav", temp_voice_pack_dir, "dash", "percentage", 5.0, -20.0
        )

        # Check the audio was generated
        assert os.path.exists(test_output_file)

    @patch.object(synthesizer.voice_manager, "get_voice_dir_path")
    @patch.object(synthesizer.voice_manager, "load_voice_config")
    def test_gma_voice_dot_list_random_choice(
        self,
        mock_load_cfg,
        mock_get_dir,
        test_output_file,
        temp_voice_pack_dir,
        dummy_audio_file_creator,
    ):
        """Test voice with 'dot' as a list of files, ensuring random.choice is used."""
        voice_name = "testvoice_list"
        dot1_path = dummy_audio_file_creator("v_dot1.wav", 100)
        dot2_path = dummy_audio_file_creator(
            "v_dot2.wav", 110
        )  # Should be chosen by our patched random.choice
        dash_path = dummy_audio_file_creator("v_dash.wav", 300)
        dot_files_list = ["v_dot1.wav", "v_dot2.wav"]

        mock_load_cfg.return_value = {
            "name": voice_name,
            "elements": {"dot": dot_files_list, "dash": "v_dash.wav"},
        }
        mock_get_dir.return_value = temp_voice_pack_dir

        # Use real files but mock random.choice to select v_dot2.wav
        with patch("random.choice", return_value="v_dot2.wav") as mock_random_choice:
            synthesizer.generate_morse_audio(
                ".", test_output_file, voice_name_full=voice_name, wpm=20
            )
            assert os.path.exists(test_output_file)

            # Check random.choice was called with the dot files list
            mock_random_choice.assert_any_call(dot_files_list)

    @patch.object(
        synthesizer.voice_manager, "load_voice_config", return_value=None
    )  # No voice
    @patch.object(synthesizer.voice_manager, "get_voice_dir_path", return_value=None)
    @patch("echomorse.synthesizer.synthesizer.generate_default_tone")
    def test_gma_default_tones_no_voice(
        self, mock_generate_default_tone, mock_get_dir, mock_load_cfg, test_output_file
    ):
        """Test generate_morse_audio uses default tones when no voice is specified."""
        # Expected durations at WPM=20
        expected_dot_duration = int(DEFAULT_UNIT_DURATION_MS * (DEFAULT_WPM / 20.0))
        expected_dash_duration = int(expected_dot_duration * 3)

        # Mock segments returned by generate_default_tone
        mock_dot = AudioSegment.silent(duration=expected_dot_duration)
        mock_dash = AudioSegment.silent(duration=expected_dash_duration)

        # Define side_effect for mock_generate_default_tone
        # It will be called for pre-caching dot, then dash, then for each symbol in "... --- ..."
        tone_map = {expected_dot_duration: mock_dot, expected_dash_duration: mock_dash}

        def generate_tone_side_effect(duration_ms, frequency=synthesizer.DEFAULT_TONE_FREQ, fade_type="percentage", fade_value=5.0):
            return tone_map.get(duration_ms, AudioSegment.silent(duration=duration_ms))

        mock_generate_default_tone.side_effect = generate_tone_side_effect

        synthesizer.generate_morse_audio(
            "... --- ...", test_output_file, voice_name_full=None, wpm=20,
            fade_type="percentage", fade_value=5.0, target_dbfs=-20.0
        )

        # Check the tone-generating function was called
        mock_generate_default_tone.assert_any_call(duration_ms=expected_dot_duration, frequency=synthesizer.DEFAULT_TONE_FREQ, fade_type="percentage", fade_value=5.0)
        mock_generate_default_tone.assert_any_call(duration_ms=expected_dash_duration, frequency=synthesizer.DEFAULT_TONE_FREQ, fade_type="percentage", fade_value=5.0)

        # Check the audio was generated
        assert os.path.exists(test_output_file)

    @patch("echomorse.synthesizer.synthesizer.AudioSegment.export")
    def test_gma_empty_or_whitespace_morse_code(self, mock_export, test_output_file):
        """Test generate_morse_audio with empty or whitespace Morse code."""
        # Prevent actual file creation by mocking export
        with pytest.raises(
            ValueError, match="Morse code cannot be empty or just whitespace."
        ):
            synthesizer.generate_morse_audio("", test_output_file)

        # Verify export wasn't called
        mock_export.assert_not_called()

        # Test with whitespace
        with pytest.raises(
            ValueError, match="Morse code cannot be empty or just whitespace."
        ):
            synthesizer.generate_morse_audio("   ", test_output_file)

        # Verify export still wasn't called
        mock_export.assert_not_called()

    def test_gma_sequence_patterns(
        self,
        temp_voice_pack_dir,
        dummy_audio_file_creator,
        voice_config_with_sequences,
        test_output_file,
    ):
        """Test sequence pattern matching in generate_morse_audio."""
        # Create the necessary audio files
        dummy_audio_file_creator("dot.wav", 100)
        dummy_audio_file_creator("dash.wav", 300)
        dummy_audio_file_creator("sos.wav", 800)  # Special SOS sound
        dummy_audio_file_creator("hello.wav", 1200)  # Special HELLO sound

        # Set up the voice manager mocks
        with patch.object(
            synthesizer.voice_manager,
            "load_voice_config",
            return_value=voice_config_with_sequences,
        ) as mock_load_cfg:
            with patch.object(
                synthesizer.voice_manager,
                "get_voice_dir_path",
                return_value=temp_voice_pack_dir,
            ) as mock_get_dir:
                # Test with pattern_chance=1.0 (always match patterns)
                synthesizer.generate_morse_audio(
                    "... --- ...",
                    test_output_file,
                    voice_name_full="test_voice_with_sequences",
                    wpm=20,
                    pattern_chance=1.0,
                )
                assert os.path.exists(test_output_file)

                # The SOS pattern should have been recognized and used
                audio = AudioSegment.from_file(test_output_file)
                # The length should be close to the sos.wav length (with slight adjustments for export/import)
                assert abs(len(audio) - 800) < 50

                # Now test with pattern_chance=0.0 (never match patterns)
                # We need a new output file to avoid overwriting the previous one
                new_output_file = test_output_file + ".no_pattern.wav"
                try:
                    synthesizer.generate_morse_audio(
                        "... --- ...",
                        new_output_file,
                        voice_name_full="test_voice_with_sequences",
                        wpm=20,
                        pattern_chance=0.0,
                    )
                    assert os.path.exists(new_output_file)

                    # The SOS pattern should NOT have been used, instead using individual dots and dashes
                    audio = AudioSegment.from_file(new_output_file)
                    # The length should be 9 symbols (3 dots, 3 dashes, 3 dots) plus spaces
                    # This would be significantly different from the single sos.wav file length
                    assert abs(len(audio) - 800) > 100
                finally:
                    # Clean up the second output file
                    if os.path.exists(new_output_file):
                        os.remove(new_output_file)

    def test_gma_with_word_and_letter_spaces(self, test_output_file):
        """Test generate_morse_audio with word spaces and letter spaces."""
        # Use default tones for simplicity
        morse_with_spaces = ". - . / ... --- ..."  # "E T E / SOS"

        synthesizer.generate_morse_audio(morse_with_spaces, test_output_file, wpm=20)
        assert os.path.exists(test_output_file)

        # Check that the audio file exists and has length > 0
        audio = AudioSegment.from_file(test_output_file)
        assert len(audio) > 0

    def test_gma_invalid_morse_char(self, test_output_file):
        """Test generate_morse_audio with invalid Morse characters."""
        # Include an invalid character '#' in the Morse code
        morse_with_invalid = "... # ..."

        with patch("logging.Logger.warning") as mock_logger:
            synthesizer.generate_morse_audio(
                morse_with_invalid, test_output_file, wpm=20
            )
            assert os.path.exists(test_output_file)

            # Check that a warning was logged for the invalid character
            mock_logger.assert_any_call(
                f"Unrecognized Morse char '#' at pos 4. Skipping."
            )

    def test_empty_final_audio_handling(self, test_output_file):
        """Test handling of an empty final audio segment (edge case)."""
        # This is a special case test where we manipulate the final_audio to be empty
        # We'll patch AudioSegment.export to check if a warning is logged
        morse_code = "."

        # Instead of trying to patch the local get_segment function directly,
        # we'll create a situation where the final audio is empty by patching
        # AudioSegment.export and verifying that the warning is logged
        with patch(
            "echomorse.synthesizer.synthesizer.AudioSegment.silent"
        ) as mock_silent:
            # First create a real silent segment for the check at the beginning
            real_silent = AudioSegment.silent(duration=0)
            # Then return a different silent segment for the 10ms fallback
            fallback_silent = AudioSegment.silent(duration=10)
            mock_silent.side_effect = [real_silent, fallback_silent]

            with patch("logging.Logger.warning") as mock_logger:
                synthesizer.generate_morse_audio(
                    morse_code, test_output_file, wpm=20,
                    fade_type="percentage", fade_value=5.0, target_dbfs=-20.0
                )
                assert os.path.exists(test_output_file)

                # Check that a warning was logged about the empty audio
                mock_logger.assert_any_call(
                    "Generated audio is empty. Creating 10ms silent file to prevent export errors."
                )

    def test_gma_cw_prosigns(
        self, temp_voice_pack_dir, dummy_audio_file_creator, test_output_file
    ):
        """Test special handling of CW prosigns with sequence patterns."""
        # Create sample audio files for testing
        dummy_audio_file_creator("dot.wav", 100)
        dummy_audio_file_creator("dash.wav", 300)
        dummy_audio_file_creator("cq.wav", 500)  # For CQ prosign
        dummy_audio_file_creator("ar.wav", 600)  # For AR (end of message) prosign
        dummy_audio_file_creator("sk.wav", 700)  # For SK (end of contact) prosign

        # Create a voice config with common CW prosigns
        cw_voice_config = {
            "name": "cw_prosigns_voice",
            "elements": {
                "dot": "dot.wav",
                "dash": "dash.wav",
                "cq": "cq.wav",
                "ar": "ar.wav",
                "sk": "sk.wav",
            },
            "sequence_patterns": {
                "cq": "-.-. --.-",  # CQ (calling any station)
                "ar": ".-.-.",  # AR (end of message)
                "sk": "...-.-",  # SK (end of contact)
            },
        }

        # Set up the voice manager mocks
        with patch.object(
            synthesizer.voice_manager, "load_voice_config", return_value=cw_voice_config
        ) as mock_load_cfg:
            with patch.object(
                synthesizer.voice_manager,
                "get_voice_dir_path",
                return_value=temp_voice_pack_dir,
            ) as mock_get_dir:
                # Test CQ pattern recognition
                synthesizer.generate_morse_audio(
                    "-.-. --.-",
                    test_output_file,  # CQ in Morse
                    voice_name_full="cw_prosigns_voice",
                    wpm=20,
                    pattern_chance=1.0,
                )
                assert os.path.exists(test_output_file)

                # The CQ pattern should have been recognized and used
                audio = AudioSegment.from_file(test_output_file)
                assert abs(len(audio) - 500) < 50  # Should be close to cq.wav length

                # Test AR pattern recognition
                ar_output_file = f"{test_output_file}.ar.wav"
                try:
                    synthesizer.generate_morse_audio(
                        ".-.-.",
                        ar_output_file,  # AR in Morse
                        voice_name_full="cw_prosigns_voice",
                        wpm=20,
                        pattern_chance=1.0,
                    )
                    assert os.path.exists(ar_output_file)

                    # The AR pattern should have been recognized and used
                    audio = AudioSegment.from_file(ar_output_file)
                    assert (
                        abs(len(audio) - 600) < 50
                    )  # Should be close to ar.wav length
                finally:
                    if os.path.exists(ar_output_file):
                        os.remove(ar_output_file)



"""
Morse Code Audio Interpreter

This module is designed to decode audio containing Morse code signals back into text.
The implementation is currently a placeholder for future development.
"""

def decode_morse_audio(audio_file_path):
    """Decode a Morse code audio file back into text (placeholder).
    
    Args:
        audio_file_path: Path to the WAV file containing Morse code audio
        
    Returns:
        str: The decoded text
    """
    # This is a placeholder for future implementation
    raise NotImplementedError("Morse audio decoding is not yet implemented.")



"""Voice manager for the Echo-Morse application.

This module handles the management of different voice types for morse code audio.
Voices are defined by a configuration file and associated audio samples.
"""
import os
import shutil
import logging
import json
import sys
import appdirs
from typing import Tuple, Optional, List, Dict, Union
from pydub import AudioSegment
from ..config import SAMPLE_RATE, DEFAULT_UNIT_DURATION_MS

logger = logging.getLogger(__name__)

# Define voice directories in priority order
VOICE_CONFIG_FILENAME = "voice_config.json"

# Primary directory: Package's audio directory (for included voices)
PACKAGE_AUDIO_DIR = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", "..", "..", "audio")
)

# User directory: OS-specific user data directory
USER_DATA_DIR = appdirs.user_data_dir("echomorse", "echomorse")
USER_AUDIO_DIR = os.path.join(USER_DATA_DIR, "audio")

# Ensure user directory exists
os.makedirs(USER_AUDIO_DIR, exist_ok=True)

# Voice directories in priority order (user dir first for overrides)
VOICE_DIRS = [USER_AUDIO_DIR, PACKAGE_AUDIO_DIR]


def get_voice_dir_path(voice_name: str) -> Optional[str]:
    """
    Gets the absolute path to a voice's directory.
    Searches both user and package directories.
    """
    if not voice_name or "/" in voice_name or "\\" in voice_name:
        return None
    
    # Search in order of priority
    for base_dir in VOICE_DIRS:
        path = os.path.join(base_dir, voice_name)
        if os.path.isdir(path) and os.path.isfile(os.path.join(path, VOICE_CONFIG_FILENAME)):
            return path
    
    return None


def load_voice_config(voice_name: str) -> Optional[Dict]:
    """Loads the voice_config.json for a given voice."""
    voice_dir = get_voice_dir_path(voice_name)
    if not voice_dir:
        return None

    config_path = os.path.join(voice_dir, VOICE_CONFIG_FILENAME)
    if not os.path.isfile(config_path):
        logger.warning(f"Voice config not found for voice '{voice_name}'")
        return None

    try:
        with open(config_path, "r") as f:
            return json.load(f)
    except Exception as e:
        logger.error(f"Error loading voice config: {e}")
        return None


def get_voice_element_samples(
    voice_name: str, element_keys: Dict[str, str] = {"dot": "dot", "dash": "dash"}
) -> Optional[Dict[str, AudioSegment]]:
    """Get audio samples for a voice."""
    if not voice_name:
        return None

    voice_config = load_voice_config(voice_name)
    if not voice_config:
        return None

    voice_dir = get_voice_dir_path(voice_name)
    if not voice_dir:
        return None

    loaded_elements: Dict[str, AudioSegment] = {}
    elements_map = voice_config.get("elements")
    if not elements_map or not isinstance(elements_map, dict):
        logger.warning(f"No 'elements' map in voice config for '{voice_name}'")
        return None

    for logical_name, config_key in element_keys.items():
        sample_filename = elements_map.get(config_key)
        if not sample_filename:
            continue

        sample_path = os.path.join(voice_dir, sample_filename)
        if not os.path.isfile(sample_path):
            continue

        try:
            segment = AudioSegment.from_file(sample_path)
            loaded_elements[logical_name] = segment
        except Exception as e:
            logger.error(f"Error loading audio file: {e}")

    return loaded_elements if loaded_elements else None


def _is_valid_voice_dir(dir_path: str) -> bool:
    """Checks if a directory contains a valid voice configuration."""
    config_path = os.path.join(dir_path, VOICE_CONFIG_FILENAME)
    return os.path.isfile(config_path)


def list_available_voices() -> List[str]:
    """List all available voices."""
    voices = ["CW (built-in)"]  # Default sine wave voice
    
    # Check all voice directories
    for base_dir in VOICE_DIRS:
        if os.path.isdir(base_dir):
            for entry_name in os.listdir(base_dir):
                entry_path = os.path.join(base_dir, entry_name)
                if (os.path.isdir(entry_path) and 
                    not entry_name.startswith(".") and
                    _is_valid_voice_dir(entry_path) and 
                    entry_name not in voices):
                    voices.append(entry_name)
    
    return voices


def install_voice(voice_name: str) -> bool:
    """
    Install a voice from the package to the user directory.
    Used for built-in voices like dog_bark.
    """
    if voice_name not in list_available_voices():
        logger.error(f"Voice '{voice_name}' not found")
        return False
    
    # Skip if already in user directory
    user_voice_path = os.path.join(USER_AUDIO_DIR, voice_name)
    if os.path.isdir(user_voice_path):
        return True
    
    # Find the voice in package directory
    package_voice_path = os.path.join(PACKAGE_AUDIO_DIR, voice_name)
    if not os.path.isdir(package_voice_path):
        logger.error(f"Built-in voice '{voice_name}' not found in package")
        return False
    
    # Copy the voice files
    try:
        shutil.copytree(package_voice_path, user_voice_path)
        logger.info(f"Installed voice '{voice_name}' to {user_voice_path}")
        return True
    except Exception as e:
        logger.error(f"Error installing voice: {e}")
        return False


def get_voice_info(voice_name: str) -> Dict:
    """Get information about a voice."""
    info = {
        "name": voice_name,
        "description": "",
        "audio_count": 0,
        "has_patterns": False,
        "location": "",
    }

    # Special case for the default CW voice
    if voice_name == "CW (built-in)":
        info["description"] = "Sine wave tones for standard Morse code"
        info["location"] = "built-in"
        return info

    # For custom voices, check their config and directory
    voice_dir = get_voice_dir_path(voice_name)
    if not voice_dir:
        return info
    
    info["location"] = voice_dir

    # Count audio files
    audio_files = 0
    for file in os.listdir(voice_dir):
        if file.lower().endswith((".wav", ".mp3", ".ogg")):
            audio_files += 1
    info["audio_count"] = audio_files

    # Get description and pattern info from config
    config = load_voice_config(voice_name)
    if config:
        if "description" in config and config["description"]:
            info["description"] = config["description"]

        # Check if the voice has custom patterns
        patterns = config.get("sequence_patterns", {})
        info["has_patterns"] = len(patterns) > 0

    return info


def print_voice_help():
    """Print help information about voices."""
    logger.info("Voice Directories:")
    for i, dir_path in enumerate(VOICE_DIRS, 1):
        exists = "✓" if os.path.isdir(dir_path) else "✗"
        logger.info(f"{i}. {dir_path} ({exists})")

    logger.info("\nAvailable Voices:")
    for voice in list_available_voices():
        info = get_voice_info(voice)
        logger.info(f"- {voice}: {info['description']}")



#!/usr/bin/env python3
"""
Command-line interface for Echo-Morse.
"""
import argparse
import sys
import os
import logging
from typing import Optional, List, Dict, Any
import io

from . import __version__
from .translator import translator
from .synthesizer import synthesizer
from .utils.voice_manager import list_available_voices, get_voice_info

# Configure logging
logger = logging.getLogger("echomorse")


def setup_logging(verbosity: int, log_file: Optional[str] = None) -> None:
    """Configure logging based on verbosity level."""
    # Set up root logger
    root_logger = logging.getLogger()

    # Clear existing handlers to avoid duplication
    for handler in root_logger.handlers[:]:
        root_logger.removeHandler(handler)

    # Set log levels based on verbosity
    if verbosity == 0:
        console_level = logging.WARNING
    elif verbosity == 1:
        console_level = logging.INFO
    else:  # verbosity >= 2
        console_level = logging.DEBUG

    # Configure console handler
    console_handler = logging.StreamHandler()
    console_handler.setLevel(console_level)
    console_formatter = logging.Formatter("%(levelname)s: %(message)s")
    console_handler.setFormatter(console_formatter)
    root_logger.addHandler(console_handler)

    # Configure file handler if log_file specified
    if log_file:
        try:
            file_handler = logging.FileHandler(log_file)
            file_handler.setLevel(
                min(console_level, logging.DEBUG)
            )  # File logs are always at least as detailed as console
            file_formatter = logging.Formatter(
                "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
            )
            file_handler.setFormatter(file_formatter)
            root_logger.addHandler(file_handler)
            logger.info(f"Logging to file: {log_file}")
        except Exception as e:
            logger.error(f"Failed to set up log file {log_file}: {e}")

    # Set root logger level to the most detailed level used by any handler
    root_logger.setLevel(min(console_level, logging.DEBUG))


def get_input_from_pipe() -> Optional[str]:
    """Check if input is being piped in and return it if available."""
    if not sys.stdin.isatty():
        return sys.stdin.read().strip()
    return None


def add_common_args(parser: argparse.ArgumentParser) -> None:
    """Add common arguments to parser."""
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase verbosity (can be used multiple times: -v, -vv)",
    )
    parser.add_argument(
        "-q", "--quiet", action="store_true", help="Suppress all output except errors"
    )
    parser.add_argument("--log-file", type=str, help="Log output to specified file")


def create_parser() -> argparse.ArgumentParser:
    """Create argument parser with subcommands."""
    # Main parser
    parser = argparse.ArgumentParser(
        prog="echomorse",
        description="Echo-Morse: Convert text to/from Morse code with customizable audio output",
    )
    add_common_args(parser)
    
    # Add version argument
    parser.add_argument(
        "--version", action="version", version=f"Echo-Morse v{__version__}"
    )

    # Create subparsers
    subparsers = parser.add_subparsers(
        dest="command", title="commands", required=True, help="Available commands"
    )

    # === TEXT TO MORSE CODE NOTATION ===
    t2c_parser = subparsers.add_parser(
        "text2code", aliases=["t2c"], help="Convert text to Morse code notation"
    )
    t2c_parser.add_argument(
        "text", nargs="?", help="Text to convert to Morse code (can also be piped in)"
    )
    t2c_parser.add_argument("-o", "--output", help="Output file (defaults to stdout)")

    # === MORSE CODE TO TEXT ===
    c2t_parser = subparsers.add_parser(
        "code2text", aliases=["c2t"], help="Convert Morse code notation to text"
    )
    c2t_parser.add_argument(
        "morse_code",
        nargs="?",
        help="Morse code to convert to text (can also be piped in)",
    )
    c2t_parser.add_argument("-o", "--output", help="Output file (defaults to stdout)")

    # === TEXT TO AUDIO ===
    t2a_parser = subparsers.add_parser(
        "text2audio", aliases=["t2a"], help="Convert text directly to Morse code audio"
    )
    t2a_parser.add_argument(
        "text", nargs="?", help="Text to convert to audio (can also be piped in)"
    )
    t2a_parser.add_argument(
        "-o",
        "--output",
        default="morse_output.wav",
        help="Output WAV file (default: morse_output.wav)",
    )
    t2a_parser.add_argument(
        "--voice",
        help="Voice profile to use (use 'list-voices' to see available options)",
    )
    t2a_parser.add_argument(
        "--wpm", type=int, default=20, help="Speed in words per minute (default: 20)"
    )
    t2a_parser.add_argument(
        "--pattern-chance",
        type=float,
        default=1.0,
        help="Chance (0.0-1.0) to use sequence patterns from voice profile (default: 1.0)",
    )
    t2a_parser.add_argument(
        "--fade-type",
        choices=["percentage", "absolute"],
        default="percentage",
        help="Type of fade to apply (default: percentage)",
    )
    t2a_parser.add_argument(
        "--fade-value",
        type=float,
        default=5.0,
        help="Fade value (percentage or milliseconds, default: 5.0)",
    )
    t2a_parser.add_argument(
        "--target-dbfs",
        type=float,
        default=-20.0,
        help="Target dBFS for audio normalization (default: -20.0)",
    )

    # === MORSE CODE TO AUDIO ===
    c2a_parser = subparsers.add_parser(
        "code2audio", aliases=["c2a"], help="Convert Morse code notation to audio"
    )
    c2a_parser.add_argument(
        "morse_code",
        nargs="?",
        help="Morse code to convert to audio (can also be piped in)",
    )
    c2a_parser.add_argument(
        "-o",
        "--output",
        default="morse_output.wav",
        help="Output WAV file (default: morse_output.wav)",
    )
    c2a_parser.add_argument(
        "--voice",
        help="Voice profile to use (use 'list-voices' to see available options)",
    )
    c2a_parser.add_argument(
        "--wpm", type=int, default=20, help="Speed in words per minute (default: 20)"
    )
    c2a_parser.add_argument(
        "--pattern-chance",
        type=float,
        default=1.0,
        help="Chance (0.0-1.0) to use sequence patterns from voice profile (default: 1.0)",
    )
    c2a_parser.add_argument(
        "--fade-type",
        choices=["percentage", "absolute"],
        default="percentage",
        help="Type of fade to apply (default: percentage)",
    )
    c2a_parser.add_argument(
        "--fade-value",
        type=float,
        default=5.0,
        help="Fade value (percentage or milliseconds, default: 5.0)",
    )
    c2a_parser.add_argument(
        "--target-dbfs",
        type=float,
        default=-20.0,
        help="Target dBFS for audio normalization (default: -20.0)",
    )

    # === AUDIO TO TEXT (placeholder) ===
    a2t_parser = subparsers.add_parser(
        "audio2text",
        aliases=["a2t"],
        help="Convert Morse code audio to text (experimental)",
    )
    a2t_parser.add_argument("audio_file", help="Audio file containing Morse code")
    a2t_parser.add_argument("-o", "--output", help="Output file (defaults to stdout)")

    # === LIST VOICES ===
    voices_parser = subparsers.add_parser(
        "list-voices", help="List available voice profiles"
    )
    voices_parser.add_argument(
        "--detailed",
        action="store_true",
        help="Show detailed information about each voice",
    )

    # === VOICE UTILITIES ===
    voice_utils_parser = subparsers.add_parser(
        "voice-utils", aliases=["voices"], help="Voice utilities and management"
    )
    voice_utils_parser.add_argument(
        "--install", metavar="VOICE_NAME",
        help="Install a built-in voice to user directory (default is all)"
    )
    voice_utils_parser.add_argument(
        "--list", action="store_true",
        help="List available voice packs and their locations"
    )

    return parser


def handle_text2code(args: argparse.Namespace) -> int:
    """Handle text to Morse code conversion."""
    # Get input from argument or stdin pipe
    text = args.text or get_input_from_pipe()
    if not text:
        logger.error("No input text provided. Use positional argument or pipe input.")
        return 1

    # Convert to Morse code
    morse_code = translator.text_to_morse(text)

    # Output result
    if args.output:
        try:
            with open(args.output, "w") as f:
                f.write(morse_code)
            logger.info(f"Morse code written to: {args.output}")
        except Exception as e:
            logger.error(f"Failed to write to output file: {e}")
            return 1
    else:
        # Print to stdout without logging prefix
        print(morse_code)

    return 0


def handle_code2text(args: argparse.Namespace) -> int:
    """Handle Morse code to text conversion."""
    # Get input from argument or stdin pipe
    morse_code = args.morse_code or get_input_from_pipe()
    if not morse_code:
        logger.error(
            "No input Morse code provided. Use positional argument or pipe input."
        )
        return 1

    # Convert to text
    text = translator.morse_to_text(morse_code)

    # Output result
    if args.output:
        try:
            with open(args.output, "w") as f:
                f.write(text)
            logger.info(f"Text written to: {args.output}")
        except Exception as e:
            logger.error(f"Failed to write to output file: {e}")
            return 1
    else:
        # Print to stdout without logging prefix
        print(text)

    return 0


def handle_text2audio(args: argparse.Namespace) -> int:
    """Handle text to audio conversion."""
    # Get input from argument or stdin pipe
    text = args.text or get_input_from_pipe()
    if not text:
        logger.error("No input text provided. Use positional argument or pipe input.")
        return 1

    # Convert text to Morse code
    morse_code = translator.text_to_morse(text)
    if not morse_code:
        logger.error("Input text did not produce valid Morse code.")
        return 1

    if args.output == "-":
        # Use a BytesIO buffer for stdout to avoid buffering issues
        buffer = io.BytesIO()
        log_input_text = None  # Avoid logging input text if streaming

        try:
            synthesizer.generate_morse_audio(
                morse_code=morse_code,
                output_file=buffer,
                voice_name_full=args.voice,
                wpm=args.wpm,
                pattern_chance=args.pattern_chance,
                fade_type=args.fade_type,
                fade_value=args.fade_value,
                target_dbfs=args.target_dbfs,
                input_text_for_log=log_input_text,
            )

            # Get the WAV data
            wav_data = buffer.getvalue()

            # Write directly to stdout file descriptor to bypass Python's stdout buffering
            os.write(1, wav_data)
            return 0

        except Exception as e:
            logger.error(f"Error generating audio: {e}")
            return 1
    else:
        # Normal file output
        output_target = args.output
        log_input_text = text  # Include input text in log for file output

        try:
            synthesizer.generate_morse_audio(
                morse_code=morse_code,
                output_file=output_target,
                voice_name_full=args.voice,
                wpm=args.wpm,
                pattern_chance=args.pattern_chance,
                fade_type=args.fade_type,
                fade_value=args.fade_value,
                target_dbfs=args.target_dbfs,
                input_text_for_log=log_input_text,
            )
            return 0
        except Exception as e:
            # Error already logged by synthesizer
            return 1


def handle_code2audio(args: argparse.Namespace) -> int:
    """Handle Morse code to audio conversion."""
    # Get input from argument or stdin pipe
    morse_code = args.morse_code or get_input_from_pipe()
    if not morse_code:
        logger.error(
            "No input Morse code provided. Use positional argument or pipe input."
        )
        return 1

    if args.output == "-":
        # Use a BytesIO buffer for stdout to avoid buffering issues
        buffer = io.BytesIO()

        try:
            synthesizer.generate_morse_audio(
                morse_code=morse_code,
                output_file=buffer,
                voice_name_full=args.voice,
                wpm=args.wpm,
                pattern_chance=args.pattern_chance,
                fade_type=args.fade_type,
                fade_value=args.fade_value,
                target_dbfs=args.target_dbfs,
                # No input_text_for_log here as we start from morse code
            )

            # Get the WAV data
            wav_data = buffer.getvalue()

            # Write directly to stdout file descriptor to bypass Python's stdout buffering
            os.write(1, wav_data)
            return 0

        except Exception as e:
            logger.error(f"Error generating audio: {e}")
            return 1
    else:
        # Normal file output
        output_target = args.output

        try:
            synthesizer.generate_morse_audio(
                morse_code=morse_code,
                output_file=output_target,
                voice_name_full=args.voice,
                wpm=args.wpm,
                pattern_chance=args.pattern_chance,
                fade_type=args.fade_type,
                fade_value=args.fade_value,
                target_dbfs=args.target_dbfs,
                # No input_text_for_log here as we start from morse code
            )
            return 0
        except Exception as e:
            # Error already logged by synthesizer
            return 1


def handle_audio2text(args: argparse.Namespace) -> int:
    """Handle audio to text conversion (placeholder)."""
    logger.warning("Audio to text conversion is not yet implemented.")
    return 1


def handle_list_voices(args: argparse.Namespace) -> int:
    """Handle listing available voices."""
    all_voices = list_available_voices()

    if not all_voices:
        logger.warning("No voices found.")
        logger.info(
            "To add a custom voice, create a subdirectory in 'audio/' (e.g., 'audio/my_voice/'),"
        )
        logger.info("add your .wav files, and a 'voice_config.json' pointing to them.")
        return 1

    # Sort voices (ensure CW is first)
    sorted_voices = sorted(all_voices, key=lambda v: (v != "CW (built-in)", v))

    # Output voice list
    print("Available voice profiles:")
    for voice_name in sorted_voices:
        voice_info = get_voice_info(voice_name)

        if args.detailed:
            # Detailed output format
            print(f"\n  {voice_name}")
            print(f"  {'=' * len(voice_name)}")
            print(f"  Description: {voice_info['description'] or 'No description'}")
            print(f"  Audio files: {voice_info['audio_count']}")
            print(f"  Has patterns: {'Yes' if voice_info['has_patterns'] else 'No'}")
        else:
            # Simple format
            description = voice_info["description"]
            desc_str = f": {description}" if description else ""
            pattern_info = " (patterns)" if voice_info["has_patterns"] else ""
            files_info = (
                f"{voice_info['audio_count']} files"
                if voice_info["audio_count"] > 0
                else "built-in"
            )

            print(f"  - {voice_name}{desc_str} ({files_info}{pattern_info})")

    return 0


def handle_voice_utils(args: argparse.Namespace) -> int:
    """Handle voice utility commands."""
    from .utils import voice_manager
    
    # Default behavior is to show basic voice info
    if not args.install and not args.list:
        print("Available voices:")
        for voice in voice_manager.list_available_voices():
            info = voice_manager.get_voice_info(voice)
            desc = f": {info['description']}" if info['description'] else ""
            print(f"  - {voice}{desc}")
        return 0
    
    # List detailed voice information
    if args.list:
        print("Voice directories:")
        for i, dir_path in enumerate([voice_manager.USER_AUDIO_DIR, voice_manager.PACKAGE_AUDIO_DIR], 1):
            exists = "✓" if os.path.isdir(dir_path) else "✗"
            print(f"{i}. {dir_path} ({exists})")
        
        print("\nDetailed voice information:")
        for voice in voice_manager.list_available_voices():
            info = voice_manager.get_voice_info(voice)
            print(f"\n  {voice}")
            print(f"  {'=' * len(voice)}")
            print(f"  Description: {info['description'] or 'None'}")
            print(f"  Location: {info['location'] or 'Unknown'}")
            print(f"  Audio files: {info['audio_count']}")
            print(f"  Has patterns: {'Yes' if info['has_patterns'] else 'No'}")
        return 0
    
    # Install voice(s)
    if args.install:
        # Install a specific voice
        voice_name = args.install
        success = voice_manager.install_voice(voice_name)
        if success:
            print(f"Successfully installed voice '{voice_name}'")
            return 0
        else:
            print(f"Failed to install voice '{voice_name}'")
            return 1
    else:
        # Install all available voices
        success_count = 0
        for voice in voice_manager.list_available_voices():
            if voice != "CW (built-in)" and voice_manager.install_voice(voice):
                success_count += 1
                print(f"Installed voice: {voice}")
        
        if success_count > 0:
            print(f"Successfully installed {success_count} voice(s)")
            return 0
        else:
            print("No voices were installed")
            return 0


def run_cli() -> int:
    """Main CLI entry point."""
    parser = create_parser()
    args = parser.parse_args()

    # Configure logging based on verbosity level
    if args.quiet:
        setup_logging(-1, args.log_file)  # -1 means only errors
    else:
        setup_logging(args.verbose, args.log_file)

    # Dispatch to appropriate handler based on command
    command_handlers = {
        "text2code": handle_text2code,
        "t2c": handle_text2code,
        "code2text": handle_code2text,
        "c2t": handle_code2text,
        "text2audio": handle_text2audio,
        "t2a": handle_text2audio,
        "code2audio": handle_code2audio,
        "c2a": handle_code2audio,
        "audio2text": handle_audio2text,
        "a2t": handle_audio2text,
        "list-voices": handle_list_voices,
        "voice-utils": handle_voice_utils,
        "voices": handle_voice_utils,
    }

    # Call the appropriate handler function
    handler = command_handlers.get(args.command)
    if handler:
        return handler(args)
    else:
        logger.error(f"Unknown command: {args.command}")
        return 1


def main() -> int:
    """Entry point for the CLI."""
    try:
        return run_cli()
    except KeyboardInterrupt:
        print("\nOperation cancelled by user", file=sys.stderr)
        return 130  # Standard exit code for SIGINT
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1


if __name__ == "__main__":
    sys.exit(main())



import pytest
import os
import sys
import subprocess
from unittest.mock import patch
from pydub import AudioSegment

# Add src directory to path for direct imports in this test file
# This might not be strictly necessary if PDM sets up the path correctly
# when running pytest, but can be a fallback.
SRC_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "../src"))
if SRC_DIR not in sys.path:
    sys.path.insert(0, SRC_DIR)

# Import translator here as it's used in a test
from echomorse.translator import translator

# Standard CW (Continuous Wave) test examples
CW_TEST_CASES = {
    "cq": "CQ",  # Calling any station
    "cq_call": "CQ CQ DE W1AW",  # Standard CQ call
    "cq_dx": "CQ DX DE JA1XYZ",  # DX (long distance) call
    "exchange": "UR RST 599 599 BK",  # Signal report exchange
    "qth": "QTH BOSTON ES NAME JOHN",  # Location and name info
    "end": "73 SK",  # End of contact (best regards, silent key)
    "prosigns": "AR BT SK",  # Common procedural signals
    "emergency": "SOS",  # Distress signal
}


class TestIntegration:
    def run_cli_command(self, command_args, capture_output=True, check=True, text_input=None):
        """Helper to run CLI commands using 'echomorse' entry point."""
        env = os.environ.copy()
        # PDM should handle the path for the 'echomorse' executable when tests are run with 'pdm run pytest'
        # Ensure the current venv is in PATH if running pytest directly without pdm run
        venv_bin_path = os.path.join(sys.prefix, "bin")
        if venv_bin_path not in env.get("PATH", ""):
             env["PATH"] = venv_bin_path + os.pathsep + env.get("PATH", "")

        # The command is now 'echomorse' followed by its arguments
        cmd = ["echomorse"] + command_args
        
        # Print the command for debugging if needed
        # print(f"\nRunning command: {' '.join(cmd)}")
        # print(f"PATH: {env.get('PATH')}")
        # print(f"PYTHONPATH: {env.get('PYTHONPATH')}")
        # print(f"Which echomorse: {subprocess.run(['which', 'echomorse'], capture_output=True, text=True, env=env).stdout}")

        process_kwargs = {
            "capture_output": capture_output,
            "text": True, # For text-based stdout/stderr
            "check": check,
            "env": env
        }
        if text_input is not None:
            process_kwargs["input"] = text_input

        return subprocess.run(cmd, **process_kwargs)

    def test_text_to_morse_cli(self):
        """Test the text-to-morse CLI command with standard CW phrases."""
        result = self.run_cli_command(["text2code", CW_TEST_CASES["cq_call"]])
        expected_morse = translator.text_to_morse(CW_TEST_CASES["cq_call"])
        assert result.returncode == 0
        assert expected_morse in result.stdout

        result = self.run_cli_command(["t2c", "SOS"]) # Test alias
        assert result.returncode == 0
        assert translator.text_to_morse("SOS") in result.stdout

    def test_morse_to_text_cli(self):
        """Test the morse-to-text CLI command with CW phrases."""
        cq_call_morse = translator.text_to_morse(CW_TEST_CASES["cq_call"])
        result = self.run_cli_command(["code2text", cq_call_morse])
        assert result.returncode == 0
        assert CW_TEST_CASES["cq_call"].upper() in result.stdout

        sos_morse = translator.text_to_morse("SOS")
        result = self.run_cli_command(["c2t", sos_morse]) # Test alias
        assert result.returncode == 0
        assert "SOS" in result.stdout

    def test_text_to_audio_cli(self, tmp_path):
        """Test text-to-morse-audio CLI command."""
        output_file = tmp_path / "t2a_output.wav"
        self.run_cli_command([
            "text2audio",
            CW_TEST_CASES["end"],
            "-o",
            str(output_file),
            "--wpm",
            "25",
        ])
        assert os.path.exists(output_file)
        # Verify file is a valid WAV file and has some content
        audio = AudioSegment.from_file(output_file)
        assert len(audio) > 100  # Check if audio has a reasonable length (e.g., >100ms)

    def test_text_to_audio_alias_cli(self, tmp_path):
        """Test text-to-morse-audio CLI command using its alias."""
        output_file = tmp_path / "t2a_alias_output.wav"
        self.run_cli_command([
            "t2a", # Using alias
            CW_TEST_CASES["exchange"],
            "-o",
            str(output_file),
            "--wpm",
            "18",
        ])
        assert os.path.exists(output_file)
        audio = AudioSegment.from_file(output_file)
        assert len(audio) > 100

    def test_list_voices_cli(self):
        """Test the list-voices CLI command."""
        result = self.run_cli_command(["list-voices"])
        assert result.returncode == 0
        assert "Available voice profiles:" in result.stdout
        assert "CW (built-in)" in result.stdout
        # Assuming dog_bark voice exists from previous tests/setup
        assert "dog_bark" in result.stdout 

    def test_list_voices_detailed_cli(self):
        """Test the list-voices --detailed CLI command."""
        result = self.run_cli_command(["list-voices", "--detailed"])
        assert result.returncode == 0
        assert "Description: Sine wave tones for standard Morse code" in result.stdout
        assert "Audio files: 0" in result.stdout # For CW (built-in)
        # Assuming dog_bark voice exists and has a config
        assert "Description: Dog bark sounds for Morse code" in result.stdout 

    def test_help_message_main(self):
        """Test that the main help message is displayed."""
        result = self.run_cli_command(["-h"], check=False) # -h exits with 0
        assert result.returncode == 0
        assert "usage: echomorse" in result.stdout
        assert "Available commands" in result.stdout

    def test_help_message_subcommand(self):
        """Test that help message for a subcommand is displayed."""
        result = self.run_cli_command(["text2audio", "-h"], check=False)
        assert result.returncode == 0
        assert "usage: echomorse text2audio" in result.stdout
        assert "Text to convert to audio" in result.stdout
    
    def test_piping_text2code_code2text(self):
        """Test piping text through text2code and then code2text."""
        input_text = "HELLO PIPING TEST"
        # Simulate: echo "HELLO PIPING TEST" | echomorse t2c | echomorse c2t
        
        # Step 1: text2code
        process1 = subprocess.Popen(["echomorse", "t2c"], stdin=subprocess.PIPE, stdout=subprocess.PIPE, text=True)
        morse_output, _ = process1.communicate(input=input_text)
        assert process1.returncode == 0
        assert morse_output.strip() == translator.text_to_morse(input_text) # Verify intermediate morse

        # Step 2: code2text
        process2 = subprocess.Popen(["echomorse", "c2t"], stdin=subprocess.PIPE, stdout=subprocess.PIPE, text=True)
        final_text, _ = process2.communicate(input=morse_output)
        assert process2.returncode == 0
        assert final_text.strip() == input_text.upper()

    def test_piping_text2code_code2audio(self, tmp_path):
        """Test piping text through text2code then code2audio."""
        output_file = tmp_path / "pipe_t2c_c2a.wav"
        input_text = "PIPE TO AUDIO"
        # Simulate: echo "PIPE TO AUDIO" | echomorse t2c | echomorse c2a -o file.wav

        process1 = subprocess.Popen(["echomorse", "t2c"], stdin=subprocess.PIPE, stdout=subprocess.PIPE, text=True)
        morse_output, _ = process1.communicate(input=input_text)
        assert process1.returncode == 0

        process2 = subprocess.Popen(["echomorse", "c2a", "-o", str(output_file)], stdin=subprocess.PIPE, stdout=subprocess.PIPE, text=True)
        _, stderr_c2a = process2.communicate(input=morse_output)
        # print(f"c2a stderr: {stderr_c2a}") # For debugging
        assert process2.returncode == 0, f"CLI Error in c2a part of pipe: {stderr_c2a}"
        
        assert os.path.exists(output_file)
        audio = AudioSegment.from_file(output_file)
        assert len(audio) > 100

    def test_text2audio_output_file_created(self, tmp_path):
        """Test text2audio command piping audio data to stdout."""
        text_to_convert = "A"
        command = [
            "echomorse", "text2audio", text_to_convert, "-o", "-",
        ]
        process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()

        assert process.returncode == 0, f"CLI Error: {stderr.decode()}"
        assert stdout.startswith(b"RIFF"), "Output does not start with RIFF (not WAV format)"
        assert stdout[8:12] == b"WAVE", "Output does not have WAVE at offset 8 (not WAV format)"
        assert len(stdout) > 1000, "Suspiciously short audio output for 'A'"

    def test_code2audio_output_file_created(self, tmp_path):
        output_file = tmp_path / "test_c2a_output.wav"
        morse_to_convert = ".... . .-.. .-.. ---"
        result = self.run_cli_command([
            "code2audio", morse_to_convert, "-o", str(output_file)
        ])
        assert result.returncode == 0, f"CLI Error: {result.stderr}"
        assert output_file.exists()
        with open(output_file, 'rb') as f:
            header = f.read(12)
        assert header.startswith(b"RIFF")
        assert header[8:12] == b"WAVE"
        assert output_file.stat().st_size > 100

    def test_text2audio_output_to_stdout(self, tmp_path):
        """Test text2audio command piping audio data to stdout."""
        text_to_convert = "A"
        command = [
            "echomorse", "text2audio", text_to_convert, "-o", "-",
        ]
        # Use subprocess.Popen to capture binary stdout
        process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()

        assert process.returncode == 0, f"CLI Error: {stderr.decode(errors='ignore')}"
        assert stdout.startswith(b"RIFF"), "Output does not start with RIFF (not WAV format)"
        assert stdout[8:12] == b"WAVE", "Output does not have WAVE at offset 8 (not WAV format)"
        assert len(stdout) > 1000, "Suspiciously short audio output for 'A' to stdout"

    def test_code2audio_output_to_stdout(self, tmp_path):
        """Test code2audio command piping audio data to stdout."""
        morse_to_convert = ".-"
        command = [
            "echomorse", "code2audio", morse_to_convert, "-o", "-",
        ]
        # Use subprocess.Popen to capture binary stdout
        process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()

        assert process.returncode == 0, f"CLI Error: {stderr.decode(errors='ignore')}"
        assert stdout.startswith(b"RIFF"), "Output does not start with RIFF (not WAV format)"
        assert stdout[8:12] == b"WAVE", "Output does not have WAVE at offset 8 (not WAV format)"
        assert len(stdout) > 1000, "Suspiciously short audio output for '.-' to stdout"

    def test_text2audio_custom_voice_and_wpm(self, tmp_path):
        # ... existing code ...
        # If the synthesizer logs a warning about voice not found, that's acceptable here.
        pass

# Any further top-level test functions or fixtures can go here



from pydub import AudioSegment

# Create a short silent segment for dot
dot_sound = AudioSegment.silent(duration=100) # 100 ms
dot_sound.export("dummy_dot.wav", format="wav")

# Create a slightly longer silent segment for dash
dash_sound = AudioSegment.silent(duration=300) # 300 ms
dash_sound.export("dummy_dash.wav", format="wav")

print("dummy_dot.wav and dummy_dash.wav created") 


import pytest
import sys
import os

# Add src directory to path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../src")))

from echomorse.translator import translator


class TestTranslator:
    def test_text_to_morse(self):
        """Test converting text to morse code."""
        # Test basic alphabets
        assert translator.text_to_morse("HELLO") == ".... . .-.. .-.. ---"
        assert translator.text_to_morse("SOS") == "... --- ..."

        # Test case insensitivity
        assert translator.text_to_morse("hello") == ".... . .-.. .-.. ---"

        # Test numbers
        assert translator.text_to_morse("123") == ".---- ..--- ...--"

        # Test punctuation
        assert translator.text_to_morse("Hi!") == ".... .. -.-.--"

        # Test mixed content
        assert (
            translator.text_to_morse("Hello, World!")
            == ".... . .-.. .-.. --- --..-- / .-- --- .-. .-.. -.. -.-.--"
        )

        # Test empty string
        assert translator.text_to_morse("") == ""

        # Test spaces
        assert translator.text_to_morse("A B") == ".- / -..."

        # Test multiple spaces
        assert translator.text_to_morse("A  B") == ".- / -..."
        assert translator.text_to_morse("A   B") == ".- / -..."

    def test_morse_to_text(self):
        """Test converting morse code to text."""
        # Test basic alphabets
        assert translator.morse_to_text(".... . .-.. .-.. ---") == "HELLO"
        assert translator.morse_to_text("... --- ...") == "SOS"

        # Test numbers
        assert translator.morse_to_text(".---- ..--- ...--") == "123"

        # Test punctuation
        assert translator.morse_to_text(".... .. -.-.--") == "HI!"

        # Test mixed content
        assert (
            translator.morse_to_text(
                ".... . .-.. .-.. --- --..-- / .-- --- .-. .-.. -.. -.-.--"
            )
            == "HELLO, WORLD!"
        )

        # Test empty string
        assert translator.morse_to_text("") == ""

        # Test invalid morse code handling (an example of what it WON'T translate if not in dict)
        assert (
            translator.morse_to_text("...... . .-.. .-.. ---") != "HELLO"
        )  # Invalid morse for 'H'

        # Test extra spaces in morse code input (should be forgiving)
        assert translator.morse_to_text("....  .  .-..  .-..  ---") == "HELLO"

        # Test leading/trailing spaces
        assert translator.morse_to_text("   .... . .-.. .-.. ---   ") == "HELLO"

    def test_roundtrip(self):
        """Test text->morse->text roundtrip.

        Ensures that translating text to morse and back results in the original text.
        This verifies the consistency of both translation functions.
        """
        original_texts = [
            "HELLO WORLD",
            "SOS",
            "123",
            "TESTING 1, 2, 3!",
            "THE QUICK BROWN FOX JUMPS OVER THE LAZY DOG",
            "HELLO, WORLD!",
            "A B C",
            # Test with prosigns
            "<AR> TEST",
            "TEST <SK>",
            "BEFORE <BT> AFTER",
            # Test with callsigns
            "CQ DE BY1QH",
            "BY1QH DE W1AW <KN>",
        ]

        for text in original_texts:
            morse = translator.text_to_morse(text)
            decoded = translator.morse_to_text(morse)
            assert decoded == text.upper(), f"Roundtrip failed for '{text}'"

    def test_cw_prosigns(self):
        """Test CW prosigns and common ham radio phrases.

        Validates both text-to-morse and morse-to-text for various CW operating
        procedures, prosigns, and conventional amateur radio exchanges.
        """
        # Test prosigns
        assert translator.text_to_morse("<AR>") == ".-.-."
        assert translator.text_to_morse("<SK>") == "...-.-"
        assert translator.text_to_morse("<BT>") == "-...-"
        assert translator.text_to_morse("<KN>") == "-.-.-"
        assert translator.text_to_morse("<AS>") == ".-..."

        # Test reverse lookup of prosigns
        assert translator.morse_to_text(".-.-.") == "<AR>"
        assert translator.morse_to_text("...-.-") == "<SK>"
        assert translator.morse_to_text("-...-") == "<BT>"
        assert translator.morse_to_text("-.-.-") == "<KN>"
        assert translator.morse_to_text(".-...") == "<AS>"

        # Test common CW abbreviations
        assert translator.text_to_morse("CQ") == "-.-. --.-"
        assert translator.text_to_morse("DE") == "-.. ."
        assert translator.text_to_morse("K") == "-.-"

        # Test full ham radio phrases
        assert (
            translator.text_to_morse("CQ CQ CQ DE W1AW W1AW K")
            == "-.-. --.- / -.-. --.- / -.-. --.- / -.. . / .-- .---- .- .-- / .-- .---- .- .-- / -.-"
        )
        assert (
            translator.text_to_morse("W1AW DE BY1QH BY1QH <KN>")
            == ".-- .---- .- .-- / -.. . / -... -.-- .---- --.- .... / -... -.-- .---- --.- .... / -.-.-"
        )
        assert translator.text_to_morse("73 <SK>") == "--... ...-- / ...-.-"

        # Test prosigns embedded within words or phrases
        msg_with_prosign = "BREAK<BT>HERE"
        morse_with_prosign = translator.text_to_morse(msg_with_prosign)
        assert morse_with_prosign == "-... .-. . .- -.- -...- .... . .-. ."

        # Test with spaces properly formatted
        proper_msg = "BREAK <BT> HERE"
        proper_morse = translator.text_to_morse(proper_msg)
        assert proper_morse == "-... .-. . .- -.- / -...- / .... . .-. ."
        assert translator.morse_to_text(proper_morse) == proper_msg

        # Test multiple prosigns in a message
        complex_msg = "TESTING <AS> WAIT <BT> CONTINUE <AR>"
        morse_complex = translator.text_to_morse(complex_msg)
        decoded = translator.morse_to_text(morse_complex)
        assert decoded == complex_msg

    def test_error_handling(self):
        """Test error handling for various edge cases."""
        # Test with non-morse characters
        assert translator.text_to_morse("αβγ") == ""  # Greek letters not in morse dict

        # Test with mix of valid and invalid chars (output depends on spacing behavior)
        special_chars_result = translator.text_to_morse("A%B@C")
        assert ".-" in special_chars_result  # A should be present
        assert "-..." in special_chars_result  # B should be present
        assert "-.-." in special_chars_result  # C should be present

        # Test with extreme input
        assert (
            translator.text_to_morse("A" * 1000).count(".") > 0
        )  # Should handle long inputs

        # Test with invalid prosigns
        assert "<XZ>" not in translator.morse_to_text(
            translator.text_to_morse("<XZ>")
        )  # Not a valid prosign
